package sim.display;
import sim.engine.*;
import sim.portrayal.*;
import java.io.*;

/** A wrapper for SimState and Schedule which provides additional functionality for
    GUI objects. This wrapper extends the functionality of SimState and
    Schedule through encapsulation, NOT subclassing.  Why?  Because the
    general idea is that SimState, and *anything* which hangs off of it,
    should be serializable out to a checkpoint and not know or care about whether
    or not it's running under a GUI or running headless.
    
    <p>Displays and Controllers (such as Display2D and the Console) do not care about the
    SimState, and make precious few calls to the Schedule.  Instead, they generally only
    work with you through the GUIState class.
    
    <p>GUIState has the same start() and finish() methods as SimState, and indeed
    the default forms of these methods just call start() and finish() on the underlying
    SimState.  Additionally, GUIState has init(Controller) and quit() methods.  The
    init(Controller) method sets up the GUIState to work in an environment controlled
    by the specified Controller.  The quit() method is called to tell the GUIState to
    shut down and free any of its resources (perhaps the simulation document is being
    closed in the GUI, or the GUI is quitting).  
    
    <p>GUIState also has two methods used by the Controller to specify things about it.
    In particular, getName() should return an intelligent name for the simulation,
    and getInfo() should return an HTML description of the simulation.
    
    <p>You can create a global inspector for your model (as opposed to the individual
    per-object inspectors created by various portrayals).  This is done by overriding either 
    getInspector() or getSimulationInspectedObject().  Additionally, override isInspectorVolatile()
    if your global inspector doesn't have to be constantly, and inefficiently, 
    updated to reflect new model information (perhaps it's just for setting initial values etcc.)
    
    <p>GUIState has a wrapper step() method which in turn calls the Schedule's
    step(state) method.  However, this wrapper provides a hook for objects such as
    displays to schedule themselves without using the Schedule.  This hook is the
    scheduleImmediate(...) and scheduleImmediateRepeat(...) methods.  There is also
    a reset() method which resets both the immediates and the underlying Schedule.
    
    <p>Last, GUIState has a wrapper function to make it convenient to read in a new SimState from
    a serialized file: readNewStateFromCheckpoint().  This function checks to see if the
    serialized file is valid for this simulation.  To do this, it calls validSimState(),
    which is a hook that you should override to return TRUE if the provided state is a valid SimState
    for your simulation (usually this means that it's the right subclass of SimState for your
    purposes).
    
    <p> Generally speaking, if you have access to a GUIState, you should use GUIstate methods
    start(), finish(), step(), reset(), and readNewStateFromCheckpoint() 
    instead of the underlying methods in the SimState and Controller.  Otherwise, feel free
    to use the underlying methods (such as Schedule.time()).
*/

public abstract class GUIState
    {
    /** The underlying SimState */
    public SimState state;
    
    /** The controller for the GUIState.  This field may be null if there is no controller
        or no controller YET */
    public Controller controller;
    
    /** Override this constructor in your subclass to call <code>super(state)</code> where state is a
        properly constructed SimState appropriate to your problem -- do NOT call <code>super()</code>*/
    private GUIState()
        {
        }
    
    /** You may optionally override this constructor to call <code>super(state)</code> but you should
        be sure to override the no-argument GUIState() constructor as stipulated. */
    protected GUIState(SimState state)
        {
        this.state = state;
        resetQueues();
        }
    
    /** Name of the simulation.  Override this to make a nicer name than "foo.bar.baz.Quux Simulation" */
    public String getName() { return "" + this.getClass().getName() + " Simulation"; }
    
    /** Returns an HTML string which gives descriptive information about the system.  The default version
        of this function looks for a file called 'index.html' located in the same directory as your
        GUIState subclass file.  If it exists and can be read, it is loaded into a string, and that string is returned.
        Otherwise an empty HTML page with a white background is returned.  Override this as you see fit.*/
    public String getInfo() 
        {
        Reader r = null;
        try
            {
            InputStream i = getClass().getResourceAsStream("index.html");
            if (i != null)
                {
                r = new BufferedReader(new InputStreamReader(i));
                StringBuffer sb = new StringBuffer();
                final int BUFLEN = 4096;
                char[] buf = new char[BUFLEN];
                int val;
                while((val = r.read(buf,0,BUFLEN)) != -1)
                    sb.append(buf,0,val);
                return sb.toString();
                }
            }
        catch (Exception e) { e.printStackTrace(); }  // hummm?
        finally { try { r.close(); } catch (Exception e) { } }
        
        return "<html><head><title></title></head><body bgcolor=\"white\"></body></html>";
        }
    
    /** Returns an Inspector which allows the user to manipulate global model features, or null if
        no such Inspector is desired.  The default returns an inspector which wraps around
        getSimulationInspectedObject(); if getSimulationInspectedObject() returns null, then getInspector()
        will return null also.  Override this as you see fit. */
    public Inspector getInspector()
        {
        Object object = getSimulationInspectedObject();
        if (object == null) return null;
        else return new SimpleInspector(object, this);
        }
    
    /** Returns an object with various property methods (getFoo(...), isFoo(...), setFoo(...)) whose
        properties will be accessible by the user.  This gives you an easy way to allow the user to
        set certain global properties of your model from the GUI.   If null is returned (the default),
        then nothing will be displayed to the user.  One trick you should know about your object:
        it should be public, as well its property methods, and if it's anonymous, it should not
        introduce any property methods not defined in its superclass.  Otherwise Java's reflection
        API can't access those methods -- they're considered private. */
    public Object getSimulationInspectedObject()
        {
        return null;
        }
        
    /** Returns true if the inspector (if any) provided in getInspector() or via getSimulationInspectedObject()
        should be updated at each time tick (as opposed to updated only when the window needs repainting).
        This is important because constant repaints are expensive.  By default, this method returns FALSE,
        which means that the inspector's updateInspector() method is called only when start() was called
        initially.  You should override this to return TRUE if your inspector needs be updated on each
        schedule tick.*/
    public boolean isInspectorVolatile()
        {
        return false;
        }
    
    /** Called to initialize (display) windows etc. 
        You can use this to set up the windows, then register them with the Controller so it can manage
        hiding, showing, and moving them.  The default version of this method simply calls
        this.controller=controller; */
    public void init(Controller controller)
        {
        this.controller = controller;
        }
    
    /** Called immediately prior to starting the simulation, or in-between
        simulation runs.  Ordinarily, you wouldn't need to override this hook. */
    public void start()
        {
        state.start();
        
        // schedule the global inspector if there is one
        
        // run the start queue stuff
        synchronized(state.schedule)
            {
            // execute the start queue
            Steppable[] _start2 = start2;
            System.arraycopy(start,0,start2,0,startSize);
            int _startSize = startSize;
            startSize = 0;
            
            // do the start stuff
            for(int x=0;x<_startSize;x++)
                _start2[x].step(state);
            }
        }
        
        
    /** Called either at the proper or a premature end to the simulation. 
        If the user quits the program, this function may not be called.  
        Ordinarily, you wouldn't need to override this hook. */
    public void finish()
        {
        synchronized(state.schedule)
            {
            // execute the finish queue
            Steppable[] _finish2 = finish2;
            System.arraycopy(finish,0,finish2,0,finishSize);
            int _finishSize = finishSize;
            finishSize = 0;
            
            // do the finish stuff
            for(int x=0;x<_finishSize;x++)
                _finish2[x].step(state);
            }

        state.finish();
        }

    /** Called by the Console when the user is quitting the SimState.  A good place
        to stick stuff that you'd ordinarily put in a finalizer -- finalizers are
        tenuous at best. So here you'd put things like the code that closes the relevant
        display windows etc.*/
    public void quit()
        {
        }

    /** This method should be set to return TRUE if state can be validly used -- mostly likely
        all you need to check is that it's the right class for this simulation.  The default
        returns TRUE if state is non-null and the same class as the current state; that's often sufficient. */
    public boolean validSimState(SimState state)
        {
        return (state != null && state.getClass().equals(this.state.getClass()));
        }

    /** Called by the Console when the user is loading in a new state from a checkpoint.  The
        new state is passed in as an argument. The default version simply sets this.state to
        the new state.  You should override this, calling super.load(state) first, 
        to reset your portrayals etc. to reflect the new state. 
        state.start() will NOT be called.  Thus anything you handled in start() that needs
        to be reset to accommodate the new state should be handled here.  We recommend that you 
        call repaint() on any Display2Ds. */
    public void load(SimState state)
        {
        this.state = state;
        }

    /** Loads a new SimState from the provided file.  Do not call this in an unthreadsafe
        situation -- it doesn't check.  Returns false if the state was not valid.  Returns
        various errors if bad things occurred trying to serialize in from the checkpoint. 
        If false is returned or an error is thrown, the old SimState is retained. */
    public boolean readNewStateFromCheckpoint(File file)
        throws IOException, ClassNotFoundException, OptionalDataException, ClassCastException, Exception
        {
        FileInputStream f = new FileInputStream(file);
        SimState state = SimState.readFromCheckpoint(f);
        f.close();
        if (!validSimState(state)) return false;
        load(state);
        return true;
        }



    // the before and after queues
    protected Steppable[] before;
    Steppable[] before2;
    protected int beforeSize;
    protected Steppable[] after;
    Steppable[] after2;
    protected int afterSize;
  
    // the start and finish queues
    protected Steppable[] start;
    Steppable[] start2;
    protected int startSize;
    protected Steppable[] finish;
    Steppable[] finish2;
    protected int finishSize;
      
    /** Don't call this unless you know what you're doing. */
    protected void resetQueues()
        {
        before = new Steppable[11];
        before2 = new Steppable[11];
        beforeSize = 0;
        after = new Steppable[11];
        after2 = new Steppable[11];
        afterSize = 0;
        start = new Steppable[11];
        start2 = new Steppable[11];
        startSize = 0;
        finish = new Steppable[11];
        finish2 = new Steppable[11];
        finishSize = 0;
        }
        
    /** Empties out the schedule and resets it to a pristine state BEFORE_SIMULATION.
        If you're using a GUIState, you should call this version instead of Schedule's
        version. */
    public synchronized final void reset(SimState state)
        {
        state.schedule.reset();
        resetQueues();
        }
    
    /** Returns FALSE if nothing was stepped -- the schedule is exhausted or time has run out. */
    public boolean step()
        {
        boolean returnval = false;
        synchronized(state.schedule)
            {
            // grab the before and after queues so no one add to them
            // while we're using them.
            Steppable[] _before2 = before2;
            Steppable[] _after2 = after2;
            System.arraycopy(before,0,before2,0,beforeSize);
            System.arraycopy(after,0,after2,0,afterSize);
            int _beforeSize = beforeSize;
            int _afterSize = afterSize;
            afterSize = 0;
            beforeSize = 0;
            
            // do the before stuff first
            for(int x=0;x<_beforeSize;x++)
                _before2[x].step(state);
            
            // step the schedule
            returnval = state.schedule.step(state);
            
            // do the after stuff
            for(int x=0;x<_afterSize;x++)
                _after2[x].step(state);
            }
        return returnval;
        }
        
    /** Roughly doubles the array size, retaining the existing elements */
    protected Steppable[] increaseSubsteps(Steppable[] substeps)
        {
        Steppable[] newsubstep = new Steppable[substeps.length*2+1];
        System.arraycopy(substeps,0,newsubstep,0,substeps.length);
        return newsubstep;
        }

    /** Schedules an item to occur (in no particular order) immediately before or immediately after
        the schedule is stepped on the next time step (not including blank steps).  
        Pass in FALSE to indicate you want to be immediately BEFORE the next timestep;
        pass in TRUE if you want to be immediately AFTER the next time step (the more common
        situation).  Returns false if the current time is AFTER_SIMULATION or if the event is null.
        
        <p>Why would you use this method?  Primarily to get things scheduled which aren't stored
        in the Schedule itself, so it can be serialized out without them. 
    */
        
    public boolean scheduleImmediate(boolean immediatelyAfter, Steppable event)
        {
        synchronized(state.schedule)
            {
            if (event == null || state.schedule.time() >= Schedule.AFTER_SIMULATION) return false;
            if (immediatelyAfter)
                {
                if (afterSize == after.length)
                    {
                    after = increaseSubsteps(after);
                    after2 = new Steppable[after.length];
                    }
                after[afterSize++] = event;
                }
            else
                {
                if (beforeSize == before.length)
                    {
                    before = increaseSubsteps(before);
                    before2 = new Steppable[before.length];
                    }
                before[beforeSize++] = event;
                }
            }
        return true;
        }
    
    /** Schedules an item to occur (in no particular order) immediately before or immediately after
        all future steps the Schedule takes (not including blank steps).  
        Pass in FALSE to indicate you want to be immediately BEFORE the next timestep;
        pass in TRUE if you want to be immediately AFTER the next time step (the more common
        situation).  Returns a Stoppable, or null if the current time is AFTER_SIMULATION or if the event is null.
        
        <p>Why would you use this method?  Primarily to get things scheduled which aren't stored
        in the Schedule itself, so it can be serialized out without them.
    */
        
    public Stoppable scheduleImmediateRepeat(boolean immediatelyAfter, Steppable event)
        {
        if (event == null) return null;
        Repeat r = new Repeat(immediatelyAfter, event);
        if (scheduleImmediate(immediatelyAfter, r)) return r;
        else return null;
        }
        
    /** Schedules an item to occur when the user starts or stops the simulator, or when it stops on its own accord.
        If atEnd is TRUE, then the item is scheduled to occur when the finish() method is executed.  If
        atEnd is FALSE, then the item is scheduled to occur when the start() method is executed.
    */
    
    public boolean scheduleAtExtreme(Steppable event, boolean atEnd)
        {
        synchronized(state.schedule)
            {
            if (event == null || state.schedule.time() >= Schedule.AFTER_SIMULATION) return false;
            if (atEnd)
                {
                if (finishSize == finish.length)
                    {
                    finish = increaseSubsteps(finish);
                    finish2 = new Steppable[finish.length];
                    }
                finish[finishSize++] = event;
                }
            else
                {
                if (startSize == start.length)
                    {
                    start = increaseSubsteps(start);
                    start2 = new Steppable[start.length];
                    }
                start[startSize++] = event;
                }
            }
        return true;
        }
        
    /** Handles repeated steps.  This is done by wrapping the Steppable with a Repeat object
        which is itself Steppable, and on its step calls its subsidiary Steppable, then reschedules
        itself.  Repeat is stopped by setting its subsidiary to null, and so the next time it's
        scheduled it won't reschedule itself (or call the subsidiary).   A private class for
        GUIState.  Notice almost exactly the same as Schedule.Repeat.  */
    
    class Repeat implements Steppable, Stoppable
        {
        protected boolean immediatelyAfter;
        protected Steppable step;  // if null, does not reschedule
        
        public Repeat(boolean immediatelyAfter, Steppable step)
            {
            this.immediatelyAfter = immediatelyAfter;
            this.step = step;
            }
        
        public synchronized void step(SimState state)
            {
            if (step!=null)
                {
                scheduleImmediate(immediatelyAfter,this);
                step.step(state);
                }
            }
        
        public synchronized void stop()  
            {
            step = null;
            }
        }
    }
    
